from abc import ABC, abstractmethod
from typing import Any, Dict, List

from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.server.feature_implementation_base import FeatureImplementationBase


class SiLAServiceBase(FeatureImplementationBase, ABC):
    @abstractmethod
    def GetFeatureDefinition(self, FeatureIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def SetServerName(self, ServerName: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> None:
        pass

    @abstractmethod
    def get_ImplementedFeatures(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> List[str]:
        pass

    @abstractmethod
    def get_ServerName(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def get_ServerType(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def get_ServerUUID(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def get_ServerDescription(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def get_ServerVersion(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass

    @abstractmethod
    def get_ServerVendorURL(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        pass



from __future__ import annotations

import logging
from collections import defaultdict
from queue import Queue
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, Optional
from uuid import UUID

from google.protobuf.message import Message
from grpc import ServicerContext

from sila2.framework.command.execution_info import CommandExecutionInfo
from sila2.framework.errors.invalid_command_execution_uuid import InvalidCommandExecutionUUID
from sila2.framework.feature import Feature
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.utils import raise_as_rpc_error
from sila2.server.feature_implementation_base import FeatureImplementationBase
from sila2.server.observables.observable_command_manager import ObservableCommandManager
from sila2.server.observables.subscription_manager_thread import SubscriptionManagerThread
from sila2.server.utils import (
    _find_allowed_defined_execution_errors,
    extract_metadata,
    raises_rpc_errors,
    unpack_parameters,
)

if TYPE_CHECKING:
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandExecutionUUID as SilaCommandExecutionUUID
    from sila2.server.sila_server import SilaServer

logger = logging.getLogger(__name__)


class FeatureImplementationServicer:
    parent_server: SilaServer
    feature: Feature
    implementation: Optional[FeatureImplementationBase]
    observable_command_managers: Dict[str, Dict[UUID, ObservableCommandManager]]
    observable_property_subscription_managers: Dict[str, SubscriptionManagerThread]

    def __init__(self, parent_server: SilaServer, feature: Feature):
        self.parent_server = parent_server
        self.feature = feature
        self.implementation = None
        self.observable_command_managers = defaultdict(dict)
        self.observable_property_subscription_managers = {}

    def _set_implementation(self, implementation: FeatureImplementationBase):
        self.implementation = implementation
        self.__start_observable_property_listeners()

        for prop in self.feature._unobservable_properties.values():
            setattr(self, f"Get_{prop._identifier}", self.__get_unobservable_property_get_call(prop._identifier))
        for cmd in self.feature._unobservable_commands.values():
            setattr(self, cmd._identifier, self.__get_unobservable_command_init_call(cmd._identifier))
        for prop in self.feature._observable_properties.values():
            setattr(
                self, f"Subscribe_{prop._identifier}", self.__get_observable_property_subscribe_call(prop._identifier)
            )
        for cmd in self.feature._observable_commands.values():
            setattr(self, cmd._identifier, self.__get_observable_command_init_call(cmd._identifier))
            setattr(self, f"{cmd._identifier}_Info", self.__get_observable_command_info_subscribe_call(cmd._identifier))
            setattr(
                self,
                f"{cmd._identifier}_Intermediate",
                self.__get_observable_command_intermediate_subscribe_call(cmd._identifier),
            )
            setattr(self, f"{cmd._identifier}_Result", self.__get_observable_command_result_get_call(cmd._identifier))
        for metadata in self.feature.metadata_definitions.values():
            setattr(
                self,
                f"Get_FCPAffectedByMetadata_{metadata._identifier}",
                self.__get_fpc_affected_by_metadata_call(metadata._identifier),
            )

    def __start_observable_property_listeners(self):
        for prop in self.feature._observable_properties.values():
            manager = SubscriptionManagerThread(
                prop.fully_qualified_identifier,
                getattr(self.implementation, f"_{prop._identifier}_producer_queue"),
                prop.to_message,
            )
            self.observable_property_subscription_managers[prop._identifier] = manager
            manager.start()

    def __get_observable_command_manager(
        self, command_id: str, execution_uuid: UUID, context: ServicerContext
    ) -> ObservableCommandManager:
        manager = self.observable_command_managers[command_id].get(execution_uuid)
        if manager is None:
            error = InvalidCommandExecutionUUID(
                f"There is no instance of Countdown with the execution uuid {execution_uuid}"
            )
            raise_as_rpc_error(error, context)
        return manager

    def __get_unobservable_command_init_call(self, command_id: str) -> Callable[[Message, ServicerContext], Message]:
        cmd = self.feature._unobservable_commands[command_id]
        impl_func: Callable = getattr(self.implementation, command_id)

        def wrapper(request: Message, context: ServicerContext) -> Message:
            metadata = extract_metadata(context, self.parent_server, cmd)
            allowed_errors = _find_allowed_defined_execution_errors(self.parent_server, cmd, metadata.keys())

            with raises_rpc_errors(context, allowed_errors):
                params = cmd.parameters.to_native_type(request)
                self.__apply_metadata_interceptors(None, metadata, cmd.fully_qualified_identifier)
                response = impl_func(*params, metadata=metadata)
                return cmd.responses.to_message(response)

        return wrapper

    def __get_unobservable_property_get_call(self, property_id: str) -> Callable[[Message, ServicerContext], Message]:
        prop = self.feature._unobservable_properties[property_id]
        impl_func: Callable = getattr(self.implementation, f"get_{property_id}")

        def wrapper(request, context: ServicerContext):
            metadata = extract_metadata(context, self.parent_server, prop)
            allowed_errors = _find_allowed_defined_execution_errors(self.parent_server, prop, metadata.keys())

            with raises_rpc_errors(context, allowed_errors):
                self.__apply_metadata_interceptors(None, metadata, prop.fully_qualified_identifier)
                response = impl_func(metadata=metadata)
                return prop.to_message(response)

        return wrapper

    def __get_observable_command_init_call(self, command_id: str) -> Callable[[Message, ServicerContext], Message]:
        cmd = self.feature._observable_commands[command_id]
        impl_func: Callable = getattr(self.implementation, command_id)

        def wrapper(request: Message, context: ServicerContext) -> Message:
            metadata = extract_metadata(context, self.parent_server, cmd)
            allowed_errors = _find_allowed_defined_execution_errors(self.parent_server, cmd, metadata.keys())

            params = unpack_parameters(cmd, request, context)

            def _func_to_execute(
                intermediate_response_queue: Queue[Any], execution_info_queue: Queue[CommandExecutionInfo]
            ) -> Message:
                with raises_rpc_errors(context, allowed_errors):
                    self.__apply_metadata_interceptors(None, metadata, cmd.fully_qualified_identifier)
                    response = impl_func(
                        *params,
                        metadata=metadata,
                        intermediate_response_queue=intermediate_response_queue,
                        execution_info_queue=execution_info_queue,
                    )
                    return cmd.responses.to_message(response)

            command_manager = ObservableCommandManager(self.parent_server, cmd, _func_to_execute, metadata.keys())
            self.observable_command_managers[command_id][command_manager.command_execution_uuid] = command_manager

            return self.feature._pb2_module.SiLAFramework__pb2.CommandConfirmation(
                commandExecutionUUID=self.feature._pb2_module.SiLAFramework__pb2.CommandExecutionUUID(
                    value=str(command_manager.command_execution_uuid)
                )
            )

        return wrapper

    def __get_observable_command_info_subscribe_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Iterator[Message]]:
        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Iterator[Message]:
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)

            with raises_rpc_errors(context):
                for value in manager.subscribe_to_execution_infos():
                    yield value

        return wrapper

    def __get_observable_command_intermediate_subscribe_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Iterator[Message]]:
        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Iterator[Message]:
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)

            with raises_rpc_errors(context):
                for value in manager.subscribe_to_intermediate_responses():
                    yield value

        return wrapper

    def __get_observable_command_result_get_call(
        self, command_id: str
    ) -> Callable[[SilaCommandExecutionUUID, ServicerContext], Message]:
        cmd = self.feature._observable_commands[command_id]

        def wrapper(request: SilaCommandExecutionUUID, context: ServicerContext) -> Message:
            manager = self.__get_observable_command_manager(command_id, UUID(request.value), context)
            allowed_errors = _find_allowed_defined_execution_errors(
                self.parent_server, cmd, manager.metadata_identifiers
            )

            with raises_rpc_errors(context, allowed_errors):
                return manager.get_responses()

        return wrapper

    def __get_observable_property_subscribe_call(
        self, property_id: str
    ) -> Callable[[Message, ServicerContext], Iterator[Message]]:
        prop = self.feature._observable_properties[property_id]
        manager = self.observable_property_subscription_managers[property_id]
        impl_func = getattr(self.implementation, f"{property_id}_on_subscription", None)

        def wrapper(request: Message, context: ServicerContext) -> Iterator[Message]:
            metadata = extract_metadata(context, self.parent_server, prop)
            allowed_errors = _find_allowed_defined_execution_errors(self.parent_server, prop, metadata.keys())

            with raises_rpc_errors(context, allowed_errors):
                self.__apply_metadata_interceptors(None, metadata, prop.fully_qualified_identifier)

                if impl_func is not None:
                    impl_func(metadata=metadata)

                for value in manager.add_subscription():
                    yield value

        return wrapper

    def __get_fpc_affected_by_metadata_call(self, metadata_id: str) -> Callable[[Message, ServicerContext], Message]:
        metadata_node = self.feature.metadata_definitions[metadata_id]
        impl_func = getattr(self.implementation, f"get_calls_affected_by_{metadata_id}")

        def wrapper(request: Message, context: ServicerContext) -> Message:
            with raises_rpc_errors(context):
                return metadata_node.to_affected_calls_message(impl_func())

        return wrapper

    def cancel_all_subscriptions(self):
        for manager in self.observable_property_subscription_managers.values():
            manager.cancel_producer()

    def __apply_metadata_interceptors(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target: FullyQualifiedIdentifier
    ):
        for interceptor in self.parent_server.metadata_interceptors:
            interceptor.intercept(parameters, metadata, target)



import os

from jinja2 import FileSystemLoader

from sila2 import resource_dir


class TemplateLoader(FileSystemLoader):
    def __init__(self):
        super().__init__(searchpath=os.path.join(resource_dir, "code_generator_templates"), encoding="utf-8")

    def get_source(self, environment, template: str):
        return super().get_source(environment, f"{template}.jinja2")



import os
from argparse import ArgumentParser
from typing import List, Optional

import sila2.code_generator
from sila2.code_generator.generator import CodeGenerator
from sila2.framework import Feature


def main(argv: Optional[List[str]] = None):
    parser = ArgumentParser(
        prog=sila2.code_generator.__name__, description="Generate Python code for given SiLA 2 feature definitions"
    )
    parser.add_argument("feature_definitions", nargs="+", help="SiLA 2 feature definitions")
    parser.add_argument("-o", "--out-dir", help="Output directory (default: '.')", default=".")

    args = parser.parse_args(argv)

    os.makedirs(args.out_dir, exist_ok=True)
    for feature_definition_file in args.feature_definitions:
        feature = Feature(open(feature_definition_file).read())
        feature_out_dir = os.path.join(args.out_dir, feature._identifier.lower())
        os.makedirs(feature_out_dir, exist_ok=True)

        generator = CodeGenerator(feature)
        generator.generate_all(feature_out_dir)
        generator.reformat_files(feature_out_dir)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from google.protobuf.message import Message

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.command.command import Command
from sila2.framework.errors.defined_execution_error import DefinedExecutionError
from sila2.framework.errors.sila_connection_error import SilaConnectionError
from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.metadata import Metadata
from sila2.framework.property.property import Property

if TYPE_CHECKING:
    from sila2.client.client_metadata import ClientMetadata
    from sila2.client.sila_client import SilaClient


def dict_to_grpc_metadata(metadata: Dict[Metadata, Any]) -> Tuple[Tuple[str, bytes], ...]:
    if metadata is None:
        return tuple()

    metadata_tuple = []
    for meta, value in metadata.items():
        metadata_tuple.append((meta.to_grpc_header_key(), meta.to_message(value)))
    return tuple(metadata_tuple)


def call_rpc_function(
    rpc_func: Any,
    parameter_message: Message,
    metadata: Optional[Dict[Union[ClientMetadata, str], Any]],
    client: SilaClient,
    origin: Optional[Union[Command, Property]],
):
    allowed_errors: List[FullyQualifiedIdentifier] = []
    metadata_dict: Dict[Metadata, Any] = {}
    if origin is not None:
        allowed_errors.extend(e.fully_qualified_identifier for e in origin.defined_execution_errors)
    if metadata is not None:
        from sila2.client.client_metadata import ClientMetadata  # avoid cyclic imports

        for key, value in metadata.items():
            if isinstance(key, ClientMetadata):
                meta = key._wrapped_metadata
            else:
                meta: Metadata = client._children_by_fully_qualified_identifier[FullyQualifiedIdentifier(key)]
            allowed_errors.extend(e.fully_qualified_identifier for e in meta.defined_execution_errors)
            metadata_dict[meta] = value

    try:
        if hasattr(rpc_func, "with_call"):
            response_msg, _ = rpc_func.with_call(parameter_message, metadata=dict_to_grpc_metadata(metadata_dict))
        else:
            response_msg = rpc_func(parameter_message, metadata=dict_to_grpc_metadata(metadata_dict))
    except Exception as ex:
        if SilaError.is_sila_error(ex):
            sila_err = SilaError.from_rpc_error(ex, client=client)
            if isinstance(sila_err, DefinedExecutionError):
                if sila_err.fully_qualified_identifier not in allowed_errors:
                    raise UndefinedExecutionError(f"{sila_err.fully_qualified_identifier}: {sila_err.message}")
                if sila_err.fully_qualified_identifier in client._registered_defined_execution_error_classes:
                    raise client._registered_defined_execution_error_classes[sila_err.fully_qualified_identifier](
                        sila_err.message
                    )
            raise sila_err
        raise SilaConnectionError(ex)

    return response_msg
